/*======================================================================

    A scp55 PCMCIA client driver

    Written by Toshiaki Nakatsu, risyu@anet.ne.jp
    
======================================================================*/

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#ifdef MODULE
#define init_scp55 init_module
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <asm/io.h>
#include <asm/system.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>
#include <drivers/sound/sound_config.h>
#ifdef PCMCIA
# include <drivers/sound/sound_calls.h>
#else
# define PCMCIA
# include <drivers/sound/sound_calls.h>
# undef PCMCIA
#endif

/*
   All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
   you do not define PCMCIA_DEBUG at all, all the debug code will be
   left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
   be present but disabled -- but it can then be enabled for specific
   modules at load time with a 'pc_debug=#' option to insmod.
*/
#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
static char *version =
"scp55_cs.c 0.10 1996/4/6 12:10:26 (Toshiaki Nakatsu)\n";
#endif

/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* Bit map of interrupts to choose from */
/* This means pick from 15, 14, 12, 11, 10, 9, 7, 5, 4, and 3 */
static u_long irq_mask = 0xdeb8;

/*====================================================================*/

/*
   The event() function is this driver's Card Services event handler.
   It will be called by Card Services when an appropriate card status
   event is received.  The config() and release() entry points are
   used to configure or release a socket, in response to card insertion
   and ejection events.  They are invoked from the scp55 event
   handler.
*/

/*static void scp55_config(u_long arg);*/
static void scp55_release(u_long arg);
static int scp55_event(event_t event, int priority,
			  event_callback_args_t *args);

/*
   The attach() and detach() entry points are used to create and destroy
   "instances" of the driver, where each instance represents everything
   needed to manage one actual PCMCIA card.
*/

static dev_link_t *scp55_attach(void);
static void scp55_detach(dev_link_t *);

/*
   You'll also need to prototype all the functions that will actually
   be used to talk to your device.  See 'pcmem_cs' for a good example
   of a fully self-sufficient driver; the other drivers rely more or
   less on other parts of the kernel.
*/

void scp55_interrupt(int reg);

/*
   The dev_info variable is the "key" that is used to match up this
   device driver with appropriate cards, through the card configuration
   database.
*/

static dev_info_t dev_info = "scp55_cs";

/*
   A linked list of "instances" of the scp55 device.  Each actual
   PCMCIA card corresponds to one device instance, and is described
   by one dev_link_t structure (defined in ds.h).

   You may not want to use a linked list for this -- for example, the
   memory card driver uses an array of dev_link_t pointers, where minor
   device numbers are used to derive the corresponding array index.
*/

static dev_link_t *dev_list = NULL;

/*
   A dev_link_t structure has fields for most things that are needed
   to keep track of a socket, but there will usually be some device
   specific information that also needs to be kept track of.  The
   'priv' pointer in a dev_link_t structure can be used to point to
   a device-specific private data structure, like this.

   A driver needs to provide a dev_node_t structure for each device
   on a card.  In some cases, there is only one device per card (for
   example, ethernet cards, modems).  In other cases, there may be
   many actual or logical devices (SCSI adapters, memory cards with
   multiple partitions).  The dev_node_t structures need to be kept
   in a linked list starting at the 'dev' field of a dev_link_t
   structure.  We allocate them in the card's private data structure,
   because they generally can't be allocated dynamically.
*/
   
typedef struct scp55_dev_t {
    dev_node_t	node;
} scp55_dev_t;

/*====================================================================*/

static void cs_error(int func, int ret)
{
    CardServices(ReportError, dev_info, (void *)func, (void *)ret);
}

/*======================================================================

    scp55_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

    The dev_link structure is initialized, but we don't actually
    configure the card at this point -- we wait until we receive a
    card insertion event.
    
======================================================================*/

static dev_link_t *scp55_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    scp55_dev_t *info;
    int ret;
    
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("scp55_attach()\n");
#endif

    /* Initialize the dev_link_t structure */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    memset(link, 0, sizeof(struct dev_link_t));
    link->release.function = &scp55_release;
    link->release.data = (u_long)link;

    /* The io structure describes IO port mapping */
    link->io.NumPorts1 = 12;
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
    link->io.NumPorts2 = 1;
    link->io.Attributes2 = IO_DATA_PATH_WIDTH_8;
    link->io.IOAddrLines = 16;

    /* Interrupt setup */
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
    link->irq.IRQInfo2 = irq_mask;
/*    link->irq.Handler = scp55_interrupt;*/
    
    /* General socket configuration */
    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    link->conf.Vpp1 = 50;
    link->conf.Vpp2 = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;
    link->conf.ConfigIndex = 1;
    link->conf.Present = PRESENT_OPTION;

    /* Allocate space for private device-specific data */
    info = kmalloc(sizeof(struct scp55_dev_t), GFP_KERNEL);
    memset(info, 0, sizeof(struct scp55_dev_t));
    link->priv = info;
    
    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
	CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
	CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
	CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &scp55_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != 0) {
	cs_error(RegisterClient, ret);
	scp55_detach(link);
	return NULL;
    }

    return link;
} /* scp55_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

static void scp55_detach(dev_link_t *link)
{
    dev_link_t **linkp;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("scp55_detach(0x%p)\n", link);
#endif
    
    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
	if (*linkp == link) break;
    if (*linkp == NULL)
	return;

    /*
       If the device is currently configured and active, we won't
       actually delete it yet.  Instead, it is marked so that when
       the release() function is called, that will trigger a proper
       detach().
    */
    if (link->state & DEV_CONFIG) {
#ifdef PCMCIA_DEBUG
        if (pc_debug)
	    printk("scp55_cs: detach postponed, '%s' still locked\n",
	       link->dev->dev_name);
#endif
	link->state |= DEV_STALE_LINK;
	return;
    }

	disable_scp55();

    /* Break the link with Card Services */
    if (link->handle)
	CardServices(DeregisterClient, link->handle);
    
    /* Unlink device structure, free pieces */
    *linkp = link->next;
    if (link->priv) {
	kfree_s(link->priv, sizeof(scp55_dev_t));
    }
    kfree_s(link, sizeof(struct dev_link_t));
    
} /* scp55_detach */

/*======================================================================

    scp55_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    ethernet device available to the system.
    
======================================================================*/

static void scp55_config(dev_link_t *link)
{
    client_handle_t handle;
    tuple_t tuple;
    cisparse_t parse;
/*    struct midi_info *info;*/
    scp55_dev_t *info;
    int i, dev;
    u_char buf[64];
    
    handle = link->handle;
    info = link->priv;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("scp55_config(0x%p)\n", link);
#endif

    /*
       This reads the card's CONFIG tuple to find its configuration
       registers.
    */
    do {
	tuple.DesiredTuple = CISTPL_CONFIG;
	i = CardServices(GetFirstTuple, handle, &tuple);
	if (i != CS_SUCCESS) break;
	tuple.TupleData = buf;
	tuple.TupleDataMax = 64;
	tuple.TupleOffset = 0;
	i = CardServices(GetTupleData, handle, &tuple);
	if (i != CS_SUCCESS) break;
	i = CardServices(ParseTuple, handle, &tuple, &parse);
	if (i != CS_SUCCESS) break;
	link->conf.ConfigBase = parse.config.base;
    } while (0);
    if (i != CS_SUCCESS) {
	cs_error(ParseTuple, i);
	link->state &= ~DEV_CONFIG_PENDING;
	return;
    }
    
    /* Configure card */
    link->state |= DEV_CONFIG;

    do {
	
        tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
        i = CardServices(GetFirstTuple, handle, &tuple);
        while (i == CS_SUCCESS) {
            i = CardServices(GetTupleData, handle, &tuple);
            if (i != CS_SUCCESS) break;
            i = CardServices(ParseTuple, handle, &tuple, &parse);
            if (i != CS_SUCCESS) break;
            if (parse.cftable_entry.io.nwin > 1) {
                link->conf.ConfigIndex = parse.cftable_entry.index;
                link->io.BasePort1 = parse.cftable_entry.io.win[0].base;
                link->io.BasePort2 = parse.cftable_entry.io.win[1].base;
                i = CardServices(RequestIO, link->handle, &link->io);
                if (i == CS_SUCCESS) break;
			}
            i = CardServices(GetNextTuple, handle, &tuple);
        }
	if (i != CS_SUCCESS) {
	    cs_error(RequestIO, i);
	    break;
	}
	
	/*
	   Now allocate an interrupt line.  Note that this does not
	   actually assign a handler to the interrupt.
	*/
	i = CardServices(RequestIRQ, link->handle, &link->irq);
	if (i != CS_SUCCESS) {
	    cs_error(RequestIRQ, i);
	    break;
	}
	
	/*
	   This actually configures the PCMCIA socket -- setting up
	   the I/O windows and the interrupt mapping.
	*/
	i = CardServices(RequestConfiguration, link->handle, &link->conf);
	if (i != CS_SUCCESS) {
	    cs_error(RequestConfiguration, i);
	    break;
	}

    } while (0);


    release_region(link->io.BasePort1, link->io.NumPorts1);/**/

	{	/* configuring soundcard by calling kernel sound driver
		 * from drivers/sound/dev_table.c:
		 *      void sound_setup (char *str, int *ints)
		 * Usually this function is summoned from init/main.c:checksetup .
		 */
	  int cardtype = SNDCARD_SCP55_MPU;
	  int ioaddr = link->io.BasePort1;
	  int irq = link->irq.AssignedIRQ;
	  int dma = 0;
	  
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	  printk("scp55_init(%d, 0x%x, %d, %d)\n",
			 cardtype,ioaddr,irq,dma);
#endif
	  dev = scp55_init(cardtype, ioaddr, irq, dma);
	  if (dev < 0) {
		printk("failed scp55_init()\n");
		scp55_release((u_long)link);
		return;
	  }
	}
	  
	/* midi_devs->info */
	/* info->dev=midi_devs[irq2dev[link->irq.AssignedIRQ]];*/
	sprintf(info->node.dev_name, "sound");
	info->node.major = SOUND_MAJOR; /*MAJOR;*/
	info->node.minor = dev;  /*SND_DEV_MIDIN;*/
	link->dev = &info->node;

    link->state &= ~DEV_CONFIG_PENDING;

    /* If any step failed, release any partially configured state */
    if (i != 0) {
	scp55_release((u_long)link);
	return;
    }

#ifdef PCMCIA_DEBUG
    if (pc_debug)
    printk("scp55 device loaded\n");
#endif
} /* scp55_config */

/*======================================================================

    After a card is removed, scp55_release() will unregister the net
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/

static void scp55_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
/*    scp55_dev_t *local = link->priv;*/

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("scp55_release(0x%p)\n", link);
#endif

    /* Unlink the device chain */
    link->dev = NULL;
    
    /* Don't bother checking to see if these succeed or not */
    CardServices(ReleaseWindow, link->win);
    CardServices(ReleaseConfiguration, link->handle);
    CardServices(ReleaseIO, link->handle, &link->io);
    CardServices(ReleaseIRQ, link->handle, &link->irq);
    link->state &= ~DEV_CONFIG;
    
    if (link->state & DEV_STALE_LINK)
	scp55_detach(link);
    
} /* scp55_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the net drivers from trying
    to talk to the card any more.

    When a CARD_REMOVAL event is received, we immediately set a flag
    to block future accesses to this device.  All the functions that
    actually access the device should check this flag to make sure
    the card is still present.
    
======================================================================*/

static int scp55_event(event_t event, int priority,
			  event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("scp55_event()\n");
#endif
    
    switch (event) {
#ifdef PCMCIA_DEBUG
    case CS_EVENT_REGISTRATION_COMPLETE:
	if (pc_debug)
	    printk("scp55_cs: registration complete\n");
	break;
#endif
    case CS_EVENT_CARD_REMOVAL:
#ifdef PCMCIA_DEBUG
	if (pc_debug)
	    printk("scp55_event: CS_EVENT_CARD_REMOVAL\n");
#endif
	link->state &= ~DEV_PRESENT;
	if (link->state & DEV_CONFIG) {
	    link->release.expires = RUN_AT(5);
	    add_timer(&link->release);
	}
	break;
    case CS_EVENT_CARD_INSERTION:
#ifdef PCMCIA_DEBUG
	if (pc_debug)
	    printk("scp55_event: CS_EVENT_CARD_INSERTION\n");
#endif		
	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	scp55_config(link);
	break;
    case CS_EVENT_PM_SUSPEND:
#ifdef PCMCIA_DEBUG
	if (pc_debug)
	    printk("scp55_event: CS_EVENT_PM_SUSPEND\n");
#endif
	link->state |= DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
#ifdef PCMCIA_DEBUG
	if (pc_debug)
	    printk("scp55_event: CS_EVENT_RESET_PHYSICAL\n");
#endif
	if (link->state & DEV_CONFIG)
	    CardServices(ReleaseConfiguration, link->handle);
	break;
    case CS_EVENT_PM_RESUME:
#ifdef PCMCIA_DEBUG
	if (pc_debug)
	    printk("scp55_event: CS_EVENT_PM_RESUME\n");
#endif
	link->state &= ~DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_CARD_RESET:
#ifdef PCMCIA_DEBUG
	if (pc_debug)
	    printk("scp55_event: CS_EVENT_CARD_RESET\n");
#endif
	if (link->state & DEV_CONFIG)
	    CardServices(RequestConfiguration, link->handle, &link->conf);
	break;
    }
    return 0;
} /* scp55_event */

/*====================================================================*/

void scp55_interrupt(int reg)
{
#ifdef PCMCIA_DEBUG
    if (pc_debug)
    printk("scp55_cs: interrupt\n");
#endif
} /* scp55_interrupt */

/*====================================================================*/

int init_scp55(void)
{
    servinfo_t serv;
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk(version);
#endif
    CardServices(GetCardServicesInfo, &serv);
    if (serv.Revision != CS_RELEASE_CODE) {
	printk("scp55: Card Services release does not match!\n");
	return -1;
    }
    register_pcmcia_driver(&dev_info, &scp55_attach, &scp55_detach);
    return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
#ifdef PCMCIA_DEBUG
    if (pc_debug)
    printk("scp55_cs: unloading\n");
#endif
    unregister_pcmcia_driver(&dev_info);
    while (dev_list != NULL) {
	if (dev_list->state & DEV_CONFIG)
	    scp55_release((u_long)dev_list);
	scp55_detach(dev_list);
    }
}
#endif /* MODULE */
